@@ -1,21 +1,20 @@
 from unittest import mock
-import unittest
 
-from cumulusci.core.config import ServiceConfig
-from cumulusci.core.config import TaskConfig
+from cumulusci.core.config import ServiceConfig, TaskConfig
 from cumulusci.tasks.github import PullRequests
 from cumulusci.tests.util import create_project_config
 
 
-class TestPullRequests(unittest.TestCase):
+class TestPullRequests:
     def test_run_task(self):
         project_config = create_project_config()
         project_config.keychain.set_service(
             "github",
+            "test_alias",
             ServiceConfig(
                 {
                     "username": "TestUser",
-                    "password": "TestPass",
+                    "token": "TestPass",
                     "email": "testuser@testdomain.com",
                 }
             ),
