@@ -1,26 +1,37 @@
-from collections import defaultdict
 import contextlib
 import functools
 import json
 import os
 import re
 import time
+from collections import defaultdict
 
+from cumulusci.core.config import BaseProjectConfig, ScratchOrgConfig, TaskConfig
+from cumulusci.core.exceptions import ProjectConfigNotFound
 from cumulusci.core.sfdx import sfdx
-from cumulusci.core.utils import process_bool_arg
-from cumulusci.core.utils import process_list_arg
-from cumulusci.tasks.salesforce import BaseRetrieveMetadata
-from cumulusci.tasks.salesforce import BaseSalesforceApiTask
+from cumulusci.core.utils import process_bool_arg, process_list_arg
 from cumulusci.tasks.metadata.package import PackageXmlGenerator
-from cumulusci.utils import temporary_dir
-from cumulusci.utils import touch
-from cumulusci.utils import inject_namespace
-from cumulusci.utils import process_text_in_directory
-from cumulusci.utils import tokenize_namespace
+from cumulusci.tasks.salesforce import BaseRetrieveMetadata, BaseSalesforceApiTask
+from cumulusci.tasks.salesforce.retrieve_profile import RetrieveProfile
+from cumulusci.utils import (
+    inject_namespace,
+    process_text_in_directory,
+    temporary_dir,
+    tokenize_namespace,
+    touch,
+)
+
+# These types are leftover from core but SFDX does not know how to process them.
+# They should be excluded from package manifests files.
+KNOWN_BAD_MD_TYPES = [
+    "AuraDefinition",
+    "ExperienceResource",
+    "LightningComponentResource",
+]
 
 
 class ListChanges(BaseSalesforceApiTask):
-    api_version = "47.0"
+    api_version = "48.0"
 
     task_options = {
         "include": {
@@ -51,18 +62,18 @@ class ListChanges(BaseSalesforceApiTask):
         self._exclude.extend(self.project_config.project__source__ignore or [])
 
     @property
-    def _snapshot_path(self):
-        parent_dir = os.path.join(".cci", "snapshot")
-        if not os.path.isdir(parent_dir):
-            os.makedirs(parent_dir)
-        return os.path.join(parent_dir, "{}.json".format(self.org_config.name))
+    @contextlib.contextmanager
+    def _snapshot_file(self):
+        with self.project_config.open_cache("snapshot") as parent_dir:
+            yield parent_dir / f"{self.org_config.name}.json"
 
     def _load_snapshot(self):
         """Load the snapshot of which component revisions have been retrieved."""
         self._snapshot = {}
-        if os.path.isfile(self._snapshot_path):
-            with open(self._snapshot_path, "r") as f:
-                self._snapshot = json.load(f)
+        with self._snapshot_file as sf:
+            if sf.exists():
+                with sf.open("r", encoding="utf-8") as f:
+                    self._snapshot = json.load(f)
 
     def _run_task(self):
         self._load_snapshot()
@@ -91,7 +102,7 @@ class ListChanges(BaseSalesforceApiTask):
     def _get_changes(self):
         """Get the SourceMember records that have changed since the last snapshot."""
         sourcemembers = self.tooling.query_all(
-            "SELECT MemberName, MemberType, RevisionNum FROM SourceMember "
+            "SELECT MemberName, MemberType, RevisionCounter FROM SourceMember "
             "WHERE IsNameObsolete=false"
         )
         changes = []
@@ -99,7 +110,7 @@ class ListChanges(BaseSalesforceApiTask):
             mdtype = sourcemember["MemberType"]
             name = sourcemember["MemberName"]
             current_revnum = self._snapshot.get(mdtype, {}).get(name)
-            new_revnum = sourcemember["RevisionNum"] or -1
+            new_revnum = sourcemember["RevisionCounter"] or -1
             if current_revnum and current_revnum == new_revnum:
                 continue
             changes.append(sourcemember)
@@ -127,32 +138,24 @@ class ListChanges(BaseSalesforceApiTask):
         for change in changes:
             mdtype = change["MemberType"]
             name = change["MemberName"]
-            revnum = change["RevisionNum"] or -1
+            revnum = change["RevisionCounter"] or -1
             self._snapshot.setdefault(mdtype, {})[name] = revnum
-        with open(self._snapshot_path, "w") as f:
-            json.dump(self._snapshot, f)
+        with self._snapshot_file as sf:
+            with sf.open("w", encoding="utf-8") as f:
+                json.dump(self._snapshot, f)
 
-    @property
-    def _maxrevision_path(self):
-        parent_dir = os.path.join(".sfdx", "orgs", self.org_config.username)
-        if not os.path.isdir(parent_dir):
-            os.makedirs(parent_dir)
-        return os.path.join(parent_dir, "maxrevision.json")
-
-    def _load_maxrevision(self):
-        """Load sfdx's maxrevision file."""
-        if not os.path.exists(self._maxrevision_path):
-            return -1
-        with open(self._maxrevision_path, "r") as f:
-            return json.load(f)
-
-    def _store_maxrevision(self, value):
-        """Update sfdx's maxrevision file."""
-        if value == -1:
-            return
-        self.logger.info(f"Setting source tracking max revision to {value}")
-        with open(self._maxrevision_path, "w") as f:
-            json.dump(value, f)
+    def _reset_sfdx_snapshot(self):
+        # If org is from sfdx, reset sfdx source tracking
+        if self.project_config.project__source_format == "sfdx" and isinstance(
+            self.org_config, ScratchOrgConfig
+        ):
+            sfdx(
+                "project reset tracking",
+                args=["-p"],
+                username=self.org_config.username,
+                capture_output=True,
+                check_return=True,
+            )
 
 
 retrieve_changes_task_options = ListChanges.task_options.copy()
@@ -166,16 +169,36 @@ retrieve_changes_task_options["api_version"] = {
         + " Defaults to project__package__api_version"
     )
 }
+retrieve_changes_task_options["retrieve_complete_profile"] = {
+    "description": (
+        "If set to True, will use RetrieveProfile to retrieve"
+        + " the complete profile. Default is set to False"
+    )
+}
 retrieve_changes_task_options["namespace_tokenize"] = BaseRetrieveMetadata.task_options[
     "namespace_tokenize"
 ]
 
+retrieve_changes_task_options["output_dir"] = {
+    "description": (
+        "The output directory for the retrieved metadata. "
+        + "If not specified, defaults to force-app or the target directory passed to retrieve changes."
+    ),
+    "required": False,
+}
+
 
 def _write_manifest(changes, path, api_version):
     """Write a package.xml for the specified changes and API version."""
     type_members = defaultdict(list)
     for change in changes:
-        type_members[change["MemberType"]].append(change["MemberName"])
+        mdtype = change["MemberType"]
+        if mdtype in KNOWN_BAD_MD_TYPES:
+            continue
+        # folders are retrieved along with their contained type
+        if mdtype.endswith("Folder"):
+            mdtype = mdtype[: -len("Folder")]
+        type_members[mdtype].append(change["MemberName"])
 
     generator = PackageXmlGenerator(
         ".",
@@ -183,10 +206,23 @@ def _write_manifest(changes, path, api_version):
         types=[MetadataType(name, members) for name, members in type_members.items()],
     )
     package_xml = generator()
-    with open(os.path.join(path, "package.xml"), "w") as f:
+    with open(os.path.join(path, "package.xml"), "w", encoding="utf-8") as f:
         f.write(package_xml)
 
 
+def separate_profiles(components):
+    """Separate the profiles from components"""
+    updated_components = []
+    profiles = []
+    for comp in components:
+        if comp["MemberType"] == "Profile":
+            profiles.append(comp["MemberName"])
+        else:
+            updated_components.append(comp)
+
+    return updated_components, profiles
+
+
 def retrieve_components(
     components,
     org_config,
@@ -195,10 +231,14 @@ def retrieve_components(
     extra_package_xml_opts: dict,
     namespace_tokenize: str,
     api_version: str,
+    project_config: BaseProjectConfig = None,
+    retrieve_complete_profile: bool = False,
+    capture_output: bool = False,
+    output_dir: str = None,
 ):
     """Retrieve specified components from an org into a target folder.
 
-    Retrieval is done using the sfdx force:source:retrieve command.
+    Retrieval is done using the sf project retrieve start command.
 
     Set `md_format` to True if retrieving into a folder with a package
     in metadata format. In this case the folder will be temporarily
@@ -207,6 +247,18 @@ def retrieve_components(
     to a namespace prefix to replace it with a `%%%NAMESPACE%%%` token.
     """
 
+    # Always use output_dir if specified, else use target
+    retrieve_target = (
+        os.path.realpath(output_dir) if output_dir else os.path.realpath(target)
+    )
+    profiles = []
+    # If retrieve_complete_profile and project_config is None, raise error
+    # This is because project_config is only required if retrieve_complete_profile is True
+    if retrieve_complete_profile and project_config is None:
+        raise ProjectConfigNotFound(
+            "Kindly provide project_config as part of retrieve_components"
+        )
+
     with contextlib.ExitStack() as stack:
         if md_format:
             # Create target if it doesn't exist
@@ -228,46 +280,68 @@ def retrieve_components(
             os.mkdir("target")
             # We need to create sfdx-project.json
             # so that sfdx will recognize force-app as a package directory.
-            with open("sfdx-project.json", "w") as f:
+            with open("sfdx-project.json", "w", encoding="utf-8") as f:
                 json.dump(
                     {"packageDirectories": [{"path": "force-app", "default": True}]}, f
                 )
             sfdx(
-                "force:mdapi:convert",
+                "project convert mdapi",
                 log_note="Converting to DX format",
                 args=["-r", target, "-d", "force-app"],
                 check_return=True,
             )
 
-        # Construct package.xml with components to retrieve, in its own tempdir
-        package_xml_path = stack.enter_context(temporary_dir(chdir=False))
-        _write_manifest(components, package_xml_path, api_version)
+        # If retrieve_complete_profile is True, separate the profiles from
+        # components to retrieve complete profile
+        if retrieve_complete_profile:
+            components, profiles = separate_profiles(components)
 
-        # Retrieve specified components in DX format
-        sfdx(
-            "force:source:retrieve",
-            access_token=org_config.access_token,
-            log_note="Retrieving components",
-            args=[
-                "-a",
-                str(api_version),
-                "-x",
-                os.path.join(package_xml_path, "package.xml"),
-                "-w",
-                "5",
-            ],
-            capture_output=False,
-            check_return=True,
-            env={"SFDX_INSTANCE_URL": org_config.instance_url},
-        )
+        if components:
+            # Construct package.xml with components to retrieve, in its own tempdir
+            package_xml_path = stack.enter_context(temporary_dir(chdir=False))
+            _write_manifest(components, package_xml_path, api_version)
 
+            # Retrieve specified components in DX format
+            p = sfdx(
+                "project retrieve start",
+                access_token=org_config.access_token,
+                log_note="Retrieving components",
+                args=[
+                    "-a",
+                    str(api_version),
+                    "-x",
+                    os.path.join(package_xml_path, "package.xml"),
+                    "-w",
+                    "5",
+                    "--ignore-conflicts",
+                    "--output-dir",
+                    retrieve_target,
+                ],
+                capture_output=capture_output,
+                check_return=True,
+                env={"SF_ORG_INSTANCE_URL": org_config.instance_url},
+            )
+
+        # Extract Profiles
+        if profiles:
+            task_config = TaskConfig(
+                config={
+                    "options": {"profiles": ",".join(profiles), "path": "force-app"}
+                }
+            )
+            cls_retrieve_profile = RetrieveProfile(
+                org_config=org_config,
+                project_config=project_config,
+                task_config=task_config,
+            )
+            cls_retrieve_profile()
         if md_format:
             # Convert back to metadata format
             sfdx(
-                "force:source:convert",
+                "project convert source",
                 log_note="Converting back to metadata format",
                 args=["-r", "force-app", "-d", target],
-                capture_output=False,
+                capture_output=capture_output,
                 check_return=True,
             )
 
@@ -286,8 +360,12 @@ def retrieve_components(
                 **extra_package_xml_opts,
             }
             package_xml = PackageXmlGenerator(**package_xml_opts)()
-            with open(os.path.join(target, "package.xml"), "w") as f:
+            with open(os.path.join(target, "package.xml"), "w", encoding="utf-8") as f:
                 f.write(package_xml)
+        if capture_output:
+            return p.stdout_text.read()
+        else:
+            return None
 
 
 class RetrieveChanges(ListChanges, BaseSalesforceApiTask):
@@ -296,12 +374,18 @@ class RetrieveChanges(ListChanges, BaseSalesforceApiTask):
     def _init_options(self, kwargs):
         super(RetrieveChanges, self)._init_options(kwargs)
         self.options["snapshot"] = process_bool_arg(kwargs.get("snapshot", True))
+        self.options["retrieve_complete_profile"] = process_bool_arg(
+            self.options.get("retrieve_complete_profile", False)
+        )
+
+        # Get output_dir first
+        output_dir = self.options.get("output_dir")
 
         # Check which directories are configured as dx packages
         package_directories = []
         default_package_directory = None
         if os.path.exists("sfdx-project.json"):
-            with open("sfdx-project.json", "r") as f:
+            with open("sfdx-project.json", "r", encoding="utf-8") as f:
                 sfdx_project = json.load(f)
                 for package_directory in sfdx_project.get("packageDirectories", []):
                     package_directories.append(package_directory["path"])
@@ -325,6 +409,7 @@ class RetrieveChanges(ListChanges, BaseSalesforceApiTask):
             md_format = path not in package_directories
         self.md_format = md_format
         self.options["path"] = path
+        self.options["output_dir"] = output_dir
 
         if "api_version" not in self.options:
             self.options[
@@ -343,6 +428,7 @@ class RetrieveChanges(ListChanges, BaseSalesforceApiTask):
             self.logger.info("{MemberType}: {MemberName}".format(**change))
 
         target = os.path.realpath(self.options["path"])
+        output_dir = self.options.get("output_dir")
         package_xml_opts = {}
         if self.options["path"] == "src":
             package_xml_opts.update(
@@ -361,19 +447,19 @@ class RetrieveChanges(ListChanges, BaseSalesforceApiTask):
             namespace_tokenize=self.options.get("namespace_tokenize"),
             api_version=self.options["api_version"],
             extra_package_xml_opts=package_xml_opts,
+            project_config=self.project_config,
+            retrieve_complete_profile=self.options["retrieve_complete_profile"],
+            output_dir=output_dir,
         )
 
         if self.options["snapshot"]:
             self.logger.info("Storing snapshot of changes")
             self._store_snapshot(filtered)
+
             if not ignored:
                 # If all changed components were retrieved,
-                # we can update the sfdx maxrevision too
-                current_maxrevision = self._load_maxrevision()
-                new_maxrevision = max(
-                    change["RevisionNum"] or -1 for change in filtered
-                )
-                self._store_maxrevision(max(current_maxrevision, new_maxrevision))
+                # we can reset sfdx source tracking too
+                self._reset_sfdx_snapshot()
 
 
 class SnapshotChanges(ListChanges):
@@ -396,8 +482,7 @@ class SnapshotChanges(ListChanges):
 
             if changes:
                 self._store_snapshot(changes)
-                maxrevision = max(change["RevisionNum"] or -1 for change in changes)
-                self._store_maxrevision(maxrevision)
+            self._reset_sfdx_snapshot()
 
     def freeze(self, step):
         return []
